
/*
 *  dus is a simple utility to display the files and directories
 *  according to their total disk occupancy.
 *
 *  Copyright (c) 2010 Francois Fleuret
 *  Written by Francois Fleuret <francois@fleuret.org>
 *
 *  This file is part of dus.
 *
 *  dus is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version 3 as
 *  published by the Free Software Foundation.
 *
 *  dus is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 *  License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with dus.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define VERSION_NUMBER "1.2"

#define _BSD_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <locale.h>
#include <getopt.h>

#define BUFFER_SIZE 4096

typedef int64_t size_sum_t;

/* Yeah, global variables! */

int ignore_dotfiles = 0; /* 1 means ignore files and directories
                            starting with a dot */

int forced_width = 0; /* -1 means no width limit, strictly positive
                         means limit, 0 means not active */

int forced_height = 0; /* -1 means no height limit, strictly positive
                           means limit, 0 means not active */

int fancy_size_display = 0; /* 1 means to use floating values with K, M and G
                               as units */

int reverse_sorting = 0; /* 1 means to show the large ones first */

int show_top = 0; /* 1 means to show the top of the sorted list
                     instead of the bottom */

size_sum_t size_min = -1; /* -1 means no minimum size, otherwise lower
                              bound on the size to display a
                              file/dir */

/********************************************************************/

/* malloc with error checking.  */

void *safe_malloc(size_t n) {
  void *p = malloc(n);
  if (!p && n != 0) {
    fprintf(stderr, "dus: Can not allocate memory: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }
  return p;
}

/********************************************************************/

int ignore_entry(const char *name) {
  return
    strcmp(name, ".") == 0 ||
    strcmp(name, "..") == 0 ||
    (ignore_dotfiles && name[0] == '.' && name[1] != '/');
}

size_sum_t entry_size(const char *name) {
  DIR *dir;
  struct dirent *dir_e;
  struct stat dummy;
  size_sum_t result;
  char subname[PATH_MAX];

  result = 0;

  if(lstat(name, &dummy) != 0) {
    fprintf(stderr, "dus: Can not stat %s: %s\n", name, strerror(errno));
    exit(EXIT_FAILURE);
  }

  if(S_ISLNK(dummy.st_mode)) {
    return 0;
  }

  if(S_ISDIR(dummy.st_mode)) {
    dir = opendir(name);
    if(dir) {
      while((dir_e = readdir(dir))) {
        if(!ignore_entry(dir_e->d_name)) {
          snprintf(subname, PATH_MAX, "%s/%s", name, dir_e->d_name);
          result += entry_size(subname);
        }
      }
      closedir(dir);
    } else {
      fprintf(stderr, "dus: Can not open directory %s: %s\n", name, strerror(errno));
      exit(EXIT_FAILURE);
    }
  } else if(S_ISREG(dummy.st_mode)) {
    result += dummy.st_size;
  }

  return result;
}

size_sum_t atoss(const char *string) {
  size_sum_t total, partial_total;
  const char *c;
  total = 0;
  partial_total = 0;

  for(c = string; *c; c++) {
    if(*c >= '0' && *c <= '9') {
      partial_total = 10 * partial_total + ((int) (*c - '0'));
    } else if(*c == 'B' || *c == 'b') {
      total += partial_total;
      partial_total = 0;
    } else if(*c == 'K' || *c == 'k') {
      total += partial_total * 1024;
      partial_total = 0;
    } else if(*c == 'M' || *c == 'm') {
      total += partial_total * 1024 * 1024;
      partial_total = 0;
    } else if(*c == 'G' || *c == 'g') {
      total += partial_total * 1024 * 1024 * 1024;
      partial_total = 0;
    } else {
      fprintf(stderr, "dus: Syntax error in size specification `%s'\n", string);
      exit(EXIT_FAILURE);
    }
  }

  total += partial_total;

  return total;
}

/**********************************************************************/

struct entry_node {
  struct entry_node *next;
  char *name;
  size_sum_t size;
};

struct entry_node *push_entry(char *name, struct entry_node *head) {
  struct entry_node *result;
  result = safe_malloc(sizeof(struct entry_node));
  result->name = strdup(name);
  result->size = entry_size(name);
  result->next = head;
  return result;
}

struct entry_node *push_dir_content(char *name, struct entry_node *head) {
  char subname[PATH_MAX];
  DIR *dir;
  struct dirent *dir_e;
  dir = opendir(name);
  if(dir) {
    while((dir_e = readdir(dir))) {
      if(!ignore_entry(dir_e->d_name)) {
        snprintf(subname, PATH_MAX, "%s/%s", name, dir_e->d_name);
        head = push_entry(subname, head);
      }
    }
    closedir(dir);
  } else {
    fprintf(stderr, "dus: Can not open directory %s: %s\n", name, strerror(errno));
    exit (EXIT_FAILURE);
  }
  return head;
}

void destroy_entry_list(struct entry_node *node) {
  struct entry_node *next;
  while(node) {
    next = node->next;
    free(node->name);
    free(node);
    node = next;
  }
}

/**********************************************************************/

int compare_files(const void *x1, const void *x2) {
  const struct entry_node **f1, **f2;

  f1 = (const struct entry_node **) x1;
  f2 = (const struct entry_node **) x2;

  if(reverse_sorting) {
    if((*f1)->size < (*f2)->size) {
      return 1;
    } else if((*f1)->size > (*f2)->size) {
      return -1;
    } else {
      return 0;
    }
  } else {
    if((*f1)->size < (*f2)->size) {
      return -1;
    } else if((*f1)->size > (*f2)->size) {
      return 1;
    } else {
      return 0;
    }
  }
}

void raw_print(char *buffer, size_t buffer_size,
               char *filename,  size_sum_t size) {
  char *a, *b, *c, u;

  b = buffer;
  do {
    if(b >= buffer + buffer_size) {
      fprintf(stderr, "dus: Buffer overflow in raw_print (hu?!).\n");
      exit(EXIT_FAILURE);
    }
    *(b++) = size%10 + '0';
    size /= 10;
  } while(size);

  a = buffer;
  c = b;
  while(a < c) {
    u = *a;
    *(a++) = *(--c);
    *c = u;
  }

  *(b++) = ' ';

  snprintf(b, buffer_size - (b - buffer), "%s\n", filename);
}

void fancy_print(char *buffer, size_t buffer_size,
                 char *filename, size_sum_t size) {
  if(size < 1024) {
    snprintf(buffer,
             buffer_size,
             "% 8d -- %s\n",
             ((int) size),
             filename);
  } else if(size < 1024 * 1024) {
    snprintf(buffer,
             buffer_size,
             "% 7.1fK -- %s\n",
             ((double) (size))/(1024.0),
             filename);
  } else if(size < 1024 * 1024 * 1024) {
    snprintf(buffer,
             buffer_size,
             "% 7.1fM -- %s\n",
             ((double) (size))/(1024.0 * 1024),
             filename);
  } else {
    snprintf(buffer,
             buffer_size,
             "% 7.1fG -- %s\n",
             ((double) (size))/(1024.0 * 1024.0 * 1024.0),
             filename);
  }
}

void print_sorted(struct entry_node *root, int width, int height) {
  char line[BUFFER_SIZE];
  struct entry_node *node;
  struct entry_node **nodes;
  int nb_nodes, n, first, last;

  nb_nodes = 0;
  for(node = root; node; node = node->next) {
    if(size_min < 0 || node->size >= size_min) {
      nb_nodes++;
    }
  }

  nodes = safe_malloc(nb_nodes * sizeof(struct entry_node *));

  n = 0;
  for(node = root; node; node = node->next) {
    if(size_min < 0 || node->size >= size_min) {
      nodes[n++] = node;
    }
  }

  qsort(nodes, nb_nodes, sizeof(struct entry_node *), compare_files);

  first = 0;
  last = nb_nodes;

  if(forced_height) {
    height = forced_height;
  }

  if(forced_width) {
    width = forced_width;
  }

  if(height >= 0 && nb_nodes > height && !show_top && !forced_height) {
    printf("...\n");
  }

  if(height > 0 && height < nb_nodes) {
    first = nb_nodes - height;
  }

  if(show_top) {
    n = last;
    last = nb_nodes - first;
    first = nb_nodes - n;
  }

  /* I do not like valgrind to complain about uninitialized data */
  if(width < BUFFER_SIZE) {
    line[width] = '\0';
  }

  for(n = first; n < last; n++) {
    if(fancy_size_display) {
      fancy_print(line, BUFFER_SIZE, nodes[n]->name, nodes[n]->size);
    } else {
      raw_print(line, BUFFER_SIZE, nodes[n]->name, nodes[n]->size);
    }
    if(width >= 1 && width + 1 < BUFFER_SIZE && line[width]) {
      line[width] = '\n';
      line[width + 1] = '\0';
    }
    printf(line);
  }

  if(height >= 0 && nb_nodes > height && show_top && !forced_height) {
    printf("...\n");
  }

  free(nodes);
}

/**********************************************************************/

void usage(FILE *out) {
  fprintf(out, "Usage: dus [OPTION]... [FILE]...\n");
  fprintf(out, "Version %s (%s)\n", VERSION_NUMBER, UNAME);
  fprintf(out, "Lists files and directories according to their size. The sizes are computed by summing recursively exact file sizes through directories. If a given directory has its name appended with '/', it is not listed, but the elements it contains are. If no files or directories are provided as arguments, the current directory is used as default.\n");
  fprintf(out, "\n");
  /*            01234567890123456789012345678901234567890123456789012345678901234567890123456789*/
  fprintf(out, "   -d, --ignore-dots          ignore files and directories starting with a '.'\n");
  fprintf(out, "   -f, --fancy                display size with float values and K, M and G\n");
  fprintf(out, "                              units.\n");
  fprintf(out, "   -r, --reverse-order        reverse the sorting order.\n");
  fprintf(out, "   -t, --show-top             show the top of the list.\n");
  fprintf(out, "   -c <cols>, --nb-columns <cols>\n");
  fprintf(out, "                              specificy the number of columns to display. The\n");
  fprintf(out, "                              value -1 corresponds to no constraint. By default\n");
  fprintf(out, "                              the command uses the tty width, or no constraint\n");
  fprintf(out, "                              if the stdout is not a tty.\n");
  fprintf(out, "   -l <lines>, --nb-lines <lines>\n");
  fprintf(out, "                              same as -c for number of lines.\n");
  fprintf(out, "   -h, --help                 show this help.\n");
  fprintf(out, "   -m <size>, --size-min <size>\n");
  fprintf(out, "                              set the listed entries minimum size. The size\n");
  fprintf(out, "                              can be specified using the G, M, K, and B units.\n");
  fprintf(out, "\n");
  fprintf(out, "Report bugs and comments to <francois@fleuret.org>.\n");
}

/**********************************************************************/

static struct option long_options[] = {
  { "ignore-dots", no_argument, 0, 'd' },
  { "reverse-order", no_argument, 0, 'r' },
  { "show-top", no_argument, 0, 't' },
  { "help", no_argument, 0, 'h' },
  { "fancy", no_argument, 0, 'f' },
  { "nb-columns", 1, 0, 'c' },
  { "nb-lines", 1, 0, 'l' },
  { "size-min", 1, 0, 'm' },
  { 0, 0, 0, 0 }
};

int main(int argc, char **argv) {
  int c, l;
  struct entry_node *root;
  struct winsize win;

  root = 0;

  setlocale (LC_ALL, "");

  while ((c = getopt_long(argc, argv, "dfrtl:c:m:hd",
                          long_options, NULL)) != -1) {
    switch (c) {

    case 'd':
      ignore_dotfiles = 1;
      break;

    case 'f':
      fancy_size_display = 1;
      break;

    case 'r':
      reverse_sorting = 1;
      break;

    case 't':
      show_top = 1;
      break;

    case 'l':
      forced_height = atoi(optarg);
      break;

    case 'c':
      forced_width = atoi(optarg);
      break;

    case 'm':
      size_min = atoss(optarg);
      break;

    case 'h':
      usage(stdout);
      exit(EXIT_SUCCESS);

      break;

    default:
      usage(stderr);
      exit(EXIT_FAILURE);
    }
  }

  if (optind < argc) {
    while (optind < argc) {
      l = strlen(argv[optind]);
      if(l > 0 && argv[optind][l - 1] == '/') {
        argv[optind][l - 1] = '\0';
        root = push_dir_content(argv[optind++], root);
      } else {
        root = push_entry(argv[optind++], root);
      }
    }
  } else {
    root = push_dir_content(".", root);
  }

  if(isatty(STDOUT_FILENO) &&
     !ioctl (STDOUT_FILENO, TIOCGWINSZ, (char *) &win)) {
    print_sorted(root, win.ws_col, win.ws_row - 2);
  } else {
    print_sorted(root, -1, -1);
  }

  destroy_entry_list(root);

  exit(EXIT_SUCCESS);
}
